-- This Source Code Form is subject to the terms of the Mozilla Public
-- License, v. 2.0. If a copy of the MPL was not distributed with this
-- file, You can obtain one at http://mozilla.org/MPL/2.0/

module AnnotationSet exposing (..)


{-| AnnotationSet aims at managing a set of annotations.
-}


import Dict
import Svg
import Html as H
import Html.Attributes as HA
import Html.Events as HE
import String
import Json.Decode as Json
import Json.Encode as JE
import List
import List.Extra as LE


import Annotation as Ann
import Utils.Helpers as HP




-- MODEL #############################################################




type alias Model_ =
    { annotations : Dict.Dict Int Ann.Model
    , selected : Maybe Int
    , uid : Int
    }


type Model = AnnSet Model_


init : (Model, Cmd msg)
init = (AnnSet <| Model_ Dict.empty Nothing 0, Cmd.none)




-- UPDATE ############################################################




type Msg
    = CreateAnnotation
    | ResetAnnotation
    | Delete
    | Select (Maybe Int)
    | Annotation Int Ann.Msg
    | Annotate Ann.Msg


update : Msg -> Model -> (Model, Cmd Msg)
update msg (AnnSet model) =
    case msg of
        CreateAnnotation ->
            let
                (annotation, _) =
                    Ann.init Nothing <| Just <| toString model.uid
                uid = model.uid
            in
                update
                    ( Select <| Just uid )
                    ( AnnSet { model
                        | annotations = Dict.insert
                            uid
                            annotation
                            model.annotations
                        , selected = Just uid
                        , uid = uid + 1
                        }
                    )
        Delete ->
            let
                annotations = case model.selected of
                    Nothing ->
                        model.annotations
                    Just id ->
                        Dict.remove id model.annotations
            in
                ( AnnSet {model | annotations = annotations }
                , Cmd.none
                )
        ResetAnnotation ->
            update
                (Annotation
                    (Maybe.withDefault -1 model.selected)
                    (Ann.Selection Nothing))
                (AnnSet model)
        Select id ->
            ( AnnSet { model | selected = id }
            , Cmd.none
            )
        Annotation id annMsg ->
            let
                maybeAnn = Dict.get id model.annotations
            in
                case maybeAnn of
                    Nothing ->
                        (AnnSet model, Cmd.none)
                    Just ann ->
                        let
                            (ann', cmdMsg) = Ann.update annMsg ann
                            annotations =
                                Dict.insert id ann' model.annotations
                        in
                            ( AnnSet {model | annotations = annotations}
                            , Cmd.map (Annotation id) cmdMsg
                            )
        Annotate annMsg ->
            case model.selected of
                Nothing ->
                    (AnnSet model, Cmd.none)
                Just id ->
                    update (Annotation id annMsg) (AnnSet model)




-- VIEW ##############################################################




selectionsView : Model -> List (Svg.Svg msg)
selectionsView (AnnSet model) =
    List.map Ann.selectionView <| Dict.values model.annotations


selectionsViewLastOnly : Model -> List (Svg.Svg msg)
selectionsViewLastOnly (AnnSet model) =
        case LE.last (Dict.values model.annotations) of
            Nothing -> []
            Just selection -> [Ann.selectionView selection]


selectHtml : Model -> H.Html Msg
selectHtml (AnnSet model) =
    H.select
        [HP.onChange <| Select << Result.toMaybe << String.toInt]
        ((H.option
            [HA.value "none", HA.selected (model.selected == Nothing)]
            [H.text "None"])
        ::
        (List.map
            (Ann.optionTag model.selected) <|
            Dict.toList model.annotations
        ))




-- OUTPUTS ##############################################################





object : Model -> JE.Value
object (AnnSet model) =
    JE.object
        [ ("annotations"
          , JE.list <| List.map Ann.object <| Dict.values model.annotations
          )
        , ("selected", case model.selected of
            Nothing -> JE.null
            Just selected -> JE.int selected
          )
        , ("uid", JE.int model.uid)
        ]


selectionsPathsObject : Model -> JE.Value
selectionsPathsObject (AnnSet model) =
    JE.list
        <| List.map Ann.selectionPathObject
        <| Dict.values model.annotations




-- OTHER #############################################################



{-| Indicates if the annotation set has at least one selection -}
hasSelection : Model -> Bool
hasSelection (AnnSet model) =
    List.any Ann.hasSelection (Dict.values model.annotations)
